-- EMIGRATION
local FilterDomeWorkplaces = function(dome, unit, only_specs)
  local dome_workplaces
  local cur_workplace = unit.workplace
  if cur_workplace and unit.lock_workplace then
    if IsKindOf(cur_workplace, "Workplace") then
      dome_workplaces = {cur_workplace}
    end
  elseif unit:CanWork() then
    local dome_dist = dome:GetOutsideWorkplacesDist()
    local avoid_workplace = unit.avoid_workplace
    if avoid_workplace and (not unit.avoid_workplace_start or unit.avoid_workplace_start + g_Consts.AvoidWorkplaceSols < unit.city.day) then
      avoid_workplace = false
    end
    local specialist = unit.specialist
    for _, workplace in ipairs(unit.city.labels.Workplace or empty_table) do
      local parent_dome = workplace.parent_dome
	  if workplace.allow_all_workers == nil then workplace.allow_all_workers = true end
      if (workplace.specialist == specialist or not only_specs) 
      and (workplace.specialist == specialist or workplace.allow_all_workers) 
      and not workplace.destroyed 
      and not workplace.demolishing 
      and (parent_dome == dome or not parent_dome and dome_dist >= HexAxialDistance(workplace, dome)) 
      and workplace.ui_working 
      and workplace:CanAcceptWorkers() 
      and workplace ~= avoid_workplace 
      and (workplace:GetFreeSlots() > 0 or workplace.specialist == specialist and workplace:GetWorkerToKick(unit, false)) 
      then
        dome_workplaces = dome_workplaces or {}
        dome_workplaces[#dome_workplaces + 1] = workplace
      end
    end
  end
  return dome_workplaces
end

function SmartFindEmigrationDome(colonist)
  local my_dome = colonist.dome
  if colonist.user_forced_dome then
    local forced_dome, forced_time = table.unpack(colonist.user_forced_dome)
    if not IsValid(forced_dome) or GameTime() - forced_time > g_Consts.ForcedByUserLockTimeout then
      colonist.user_forced_dome = false
    elseif forced_dome == my_dome then
      return
    else
      local is_walking, dist = IsInWalkingDist(forced_dome, my_dome)
      return forced_dome, is_walking, dist
    end
  end
  local traits = colonist.traits
  local is_child = traits.Child
  local eval = -1
  local pos = my_dome or colonist
  local LR_transport_available = IsLRTransportAvailable(colonist.city)
  local choosen, spec_workplaces, non_spec_workplaces
  if my_dome then
    if not my_dome.accept_colonists then
      return
    elseif not colonist:IsHomeless() and (my_dome.ui_working or GameTime() - my_dome.ui_working_changed > const.ColonistMaxWaitInDomeTurnedOff) then
      eval = TraitFilterColonist(my_dome.traits_filter, traits)
    end
    spec_workplaces = FilterDomeWorkplaces(my_dome, colonist, true) --GetFreeWorkplacesAround(my_dome) > 0
    non_spec_workplaces = FilterDomeWorkplaces(my_dome, colonist, false)
  end
  local is_unemployed = colonist:CanWork() and not IsValid(colonist.workplace) and not colonist.user_forced_workplace
  local is_walking, dist
  for _, dome in ipairs(colonist.city.labels.Dome or empty_table) do
    if my_dome ~= dome and not dome.destroyed and not dome.demolishing and dome.accept_colonists and dome.ui_working and dome:HasPower() and dome:HasWater() and dome:HasAir() then
      local new_is_walking, new_dist = IsInWalkingDist(dome, pos)
      if new_is_walking or LR_transport_available then
        local new_eval = TraitFilterColonist(dome.traits_filter, traits)
        local free_living_space, free_living_space_adult = dome:GetFreeLivingSpace(is_child)
        
        local new_spec_workplaces
        local new_nonspec_workplaces 
        if new_eval >= 0 then
          new_spec_workplaces = FilterDomeWorkplaces(dome, colonist, true) --GetFreeWorkplacesAround(dome) > 0
          new_nonspec_workplaces = FilterDomeWorkplaces(dome, colonist, false)
        end
        
        if new_eval >= 0 and eval <= new_eval and (
          (
            colonist:CanWork() and Random(0, 100) < 10 and 
            (
              (IsValid(colonist.workplace) and not colonist.user_forced_workplace and 
              colonist.workplace.specialist ~= colonist.specialist and not spec_workplaces and new_spec_workplaces)
              or
              (not IsValid(colonist.workplace) and not spec_workplaces and new_spec_workplaces) 
            )
          )
          or
          ( 
            (colonist:IsHomeless() or (is_unemployed and not (spec_workplaces or non_spec_workplaces))) 
            and 
            (new_spec_workplaces or new_nonspec_workplaces) 
            and 
            (is_child and free_living_space > 0 or not is_child and free_living_space_adult > 0) 
          )
          or
          (not colonist:CanWork() and eval < new_eval and (is_child and free_living_space > 0 or not is_child and free_living_space_adult > 0))
        ) then
          choosen = dome
          eval = new_eval
          is_walking = new_is_walking
          dist = new_dist
        end
      end
    end
  end
  return choosen, is_walking, dist
end

VanillaFindEmigrationDome = Colonist.FindEmigrationDome
Colonist.FindEmigrationDome = SmartFindEmigrationDome